from .abstractuser import AbstractUser
from .utils.converter import Converter
from .utils.wrap_tools import make_repr

class UserStats(AbstractUser):
    """Class that extends :class:`.AbstractUser`, adding
    user's statistics to it.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self.options = options

    def __repr__(self):
        info = {
            'account_id': self.account_id,
            'name': self.name,
            'id': self.id,
            'lb_place': Converter.to_ordinal(self.lb_place),
            'stars': self.stars,
            'demons': self.demons,
            'cp': self.cp
        }
        return make_repr(self, info)

    @property
    def stars(self):
        """:class:`int`: Amount of stars the user has."""
        return self.options.get('stars')

    @property
    def demons(self):
        """:class:`int`: Amount of demons the user has beaten."""
        return self.options.get('demons')

    @property
    def cp(self):
        """:class:`int`: Amount of Creator Points the user has."""
        return self.options.get('cp')

    @property
    def diamonds(self):
        """:class:`int`: Amount of diamonds the user has."""
        return self.options.get('diamonds')

    @property
    def coins(self):
        """:class:`int`: Number of coins the user has."""
        return self.options.get('coins')

    @property
    def user_coins(self):
        """:class:`int`: Amount of User Coins user has."""
        return self.options.get('user_coins')

    @property
    def lb_place(self):
        """:class:`int`: User's place in leaderboard. ``-1`` if not set."""
        return self.options.get('lb_place', -1)

    def has_cp(self):
        """:class:`bool`: Indicates if a user has Creator Points."""
        return self.cp > 0

    async def update(self):
        """|coro|

        Update ``self``.
        """
        new = await self._client.fetch_user(self.account_id, stats=True)
        self.options = new.options

class User(UserStats):
    """Class that represents a Geometry Dash User.
    This class is derived from :class:`.UserStats`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self.options = options

    def __repr__(self):
        info = {
            'account_id': self.account_id,
            'id': self.id,
            'name': self.name,
            'role': self.role,
            'cp': self.cp
        }
        return make_repr(self, info)

    @property
    def role(self):
        """:class:`.StatusLevel`: A status level of the user."""
        return self.options.get('role')

    @property
    def rank(self):
        """Union[:class:`int`, ``None``]: A global rank of the user.
        ``None`` if the user is not on the leaderboard.
        """
        return self.options.get('global_rank')

    @property
    def youtube(self):
        """Union[:class:`str`, ``None``]: A youtube name of the user."""
        return self.options.get('youtube')[0]

    @property
    def youtube_link(self):
        """Union[:class:`str`, ``None``]: A link to the user's youtube channel."""
        return self.options.get('youtube')[1]

    @property
    def twitter(self):
        """Union[:class:`str`, ``None``]: A twitter name of the user."""
        return self.options.get('twitter')[0]

    @property
    def twitter_link(self):
        """Union[:class:`str`, ``None``]: A link to the user's twitter page."""
        return self.options.get('twitter')[1]

    @property
    def twitch(self):
        """Union[:class:`str`, ``None``]: A twitch name of the user."""
        return self.options.get('twitch')[0]

    @property
    def twitch_link(self):
        """Union[:class:`str`, ``None``]: A link to the user's twitch channel."""
        return self.options.get('twitch')[1]

    @property
    def msg_policy(self):
        """:class:`.MessagePolicyType`: A type indicating user's message inbox policy."""
        return self.options.get('messages')

    @property
    def friend_req_policy(self):
        """:class:`.FriendRequestPolicyType`: A type indicating user's friend requests policy."""
        return self.options.get('friend_requests')

    @property
    def comments_policy(self):
        """:class:`.CommentPolicyType`: A type indicating user's comment history policy."""
        return self.options.get('comments')

    @property
    def icon_set(self):
        """:class:`.IconSet`: An iconset of the user."""
        return self.options.get('icon_setup')

    def is_mod(self, elder: str = None):
        """:class:`bool`: Indicates if a user is Geometry Dash (Elder) Moderator.
        For instance, *RobTop* is an Elder Moderator, that means:
        ``robtop.is_mod() -> True`` and ``robtop.is_mod('elder') -> True``.
        """
        if self.role is None:
            return False

        elif elder is None:
            return self.role.value >= 1

        elif elder == 'elder':
            return self.role.value == 2

        raise TypeError("is_mod(elder) expected elder=='elder', or None.")

    async def update(self):
        """|coro|

        Update the user's statistics and other parameters.
        """
        new = await self._client.get_user(self.account_id)
        self.options = new.options



from .enums import DemonDifficulty, LevelDifficulty

from ..song import Song

class Converter:
    """Some weird class where NeKit holds his converters for everything"""
    @classmethod
    def to_normal_song(cls, song_id: int):
        cases = {
            0: ("ForeverBound", "Stereo Madness"),
            1: ("DJVI", "Back On Track"),
            2: ("Step", "Polargeist"),
            3: ("DJVI", "Dry Out"),
            4: ("DJVI", "Base After Base"),
            5: ("DJVI", "Cant Let Go"),
            6: ("Waterflame", "Jumper"),
            7: ("Waterflame", "Time Machine"),
            8: ("DJVI", "Cycles"),
            9: ("DJVI", "xStep"),
            10: ("Waterflame", "Clutterfunk"),
            11: ("DJ-Nate", "Theory of Everything"),
            12: ("Waterflame", "Electroman Adventures"),
            13: ("DJ-Nate", "Clubstep"),
            14: ("DJ-Nate", "Electrodynamix"),
            15: ("Waterflame", "Hexagon Force"),
            16: ("Waterflame", "Blast Processing"),
            17: ("DJ-Nate", "Theory of Everything 2"),
            18: ("Waterflame", "Geometrical Dominator"),
            19: ("F-777", "Deadlocked"),
            20: ("MDK", "Fingerdash"),
        }
        author, name = cases.get(song_id, (None, None))
        return Song(
            name = name, author = author, id = song_id,
            size = None, links = (None, None), custom = False
        )

    @classmethod
    def to_ordinal(cls, n: int):
        x = str(n)

        pn = x[-2:-1]
        sn = x[-1:]

        cases = {
            '1': 'st',
            '2': 'nd',
            '3': 'rd'
        }

        res = x + (
            cases.get(sn, 'th') if pn != '1' else 'th'
        )
        return res

    @classmethod
    def value_to_pack_difficulty(cls, value: int):
        cases = {
            1: LevelDifficulty.EASY,
            2: LevelDifficulty.NORMAL,
            3: LevelDifficulty.HARD,
            4: LevelDifficulty.HARDER,
            5: LevelDifficulty.INSANE,
            6: LevelDifficulty.DEMON
        }
        return cases.get(value, LevelDifficulty.NA)

    @classmethod
    def value_to_difficulty(cls, value: int):
        cases = {
            10: LevelDifficulty.EASY,
            20: LevelDifficulty.NORMAL,
            30: LevelDifficulty.HARD,
            40: LevelDifficulty.HARDER,
            50: LevelDifficulty.INSANE
        }
        return cases.get(value, LevelDifficulty.NA)

    @classmethod
    def value_to_demon(cls, value: int):
        cases = {
            3: DemonDifficulty.EASY_DEMON,
            4: DemonDifficulty.MEDIUM_DEMON,
            5: DemonDifficulty.INSANE_DEMON,
            6: DemonDifficulty.EXTREME_DEMON
        }
        return cases.get(value, DemonDifficulty.HARD_DEMON)

    @classmethod
    def convert_level_difficulty(cls, diff: int, demon_diff: int, is_demon: bool, is_auto: bool):
        if is_auto:
            return LevelDifficulty.AUTO
        if is_demon:
            return cls.value_to_demon(demon_diff)
        return cls.value_to_difficulty(diff)



from operator import attrgetter as attrget

def find(predicate, seq, *, _all: bool = False):
    """For each element in sequence, return first element if predicate
    returns ``True`` and ``'_all'`` is ``False``.

    Otherwise, find all elements matching and return them.

    Example:

    .. code-block:: python3

        ...
        friends = await client.get_friends()
        old_users = gd.utils.find(lambda x: x.account_id < 500_000, friends, _all=True)

    """
    res = []
    for elem in seq:
        if predicate(elem):
            if not _all:
                return elem
            res.append(elem)

    if _all:
        return res


def get(iterable, **attrs):
    """For each element in iterable, return first element that matches
    requirements and ``'_all'`` is ``False``.

    Otherwise, find all elements matching and return them.

    Example:

    .. code-block:: python3

        ...
        friends = await client.get_friends()
        nekit = gd.utils.get(friends, name='NeKitDS')

    """
    # check if ALL elements matching requirements should be returned
    _all = attrs.pop("_all", False)

    converted = [
        (attrget(attr.replace('__', '.')), value)
        for attr, value in attrs.items()
    ]

    res = []
    for elem in iterable:
        if all(pred(elem) == value for pred, value in converted):
            if not _all:
                return elem
            res.append(elem)

    if _all:
        return res



from .generator import SpriteGenUtil, SpriteGen, test_path, gen_center_offset, fix_glow_color, predict_color, reorder
from ..colors import colors
from ..iconset import IconSet
import time

gco = gen_center_offset
sg = SpriteGen()
sgu = SpriteGenUtil()
img = sg.load_image()
full_test = [
    IconSet(color_1 = colors[10], color_2 = colors[3], icon_robot = i) for i in range(1,35)
]
parsed = sgu.parse_xml()
full_bases = [
    sgu.find_bases(iconset) for iconset in full_test
]
full_related = [
    sgu.format_related(parsed, base_list) for base_list in full_bases
]
full_sprites = [
    sgu.retrieve_sprites(parsed, 'robot', elem) for elem in full_related
]
def make_sprite(type, splist, iconset) -> None:
    reorder(splist)
    s = splist
    if not iconset.has_glow_outline():
        s = [sp for sp in splist if '_glow_' not in sp.name]
    ret = sg.make_blank()
    colors = iconset.get_colors()
    for sprite in s:
        print(sprite)
        color = predict_color(sprite, colors)
        sp = sg.get_sprite_image(img, sprite)
        offset = gco(ret, sp, sprite)
        if color is not None:
            sp = sg.color(sp, color)
        ret.alpha_composite(sp, offset)
        ret.save(test_path)
        time.sleep(1)

for i in range(34):
    make_sprite('robot', full_sprites[i], full_test[i])



import asyncio
import logging
from typing import Callable, Dict

import aiohttp

from ..errors import HTTPNotConnected
from .mapper import mapper_util
from .wrap_tools import make_repr

log = logging.getLogger(__name__)

class HTTPClient:
    """Class that handles the main part of the entire gd.py - sending HTTP requests.

    Attributes
    ----------
    semaphore: Optional[:class:`asyncio.Semaphore`]
        A semaphore to use when doing requests. Defaults to :class:`asyncio.Semaphore` with value ``200``.
    """
    BASE = 'http://www.boomlings.com/database/'
    VALID_ERRORS = (
        OSError,
        aiohttp.ClientError
    )

    def __init__(self, *, semaphore=None):
        self.semaphore = semaphore or asyncio.Semaphore(250)
        self.debug = False

    def __repr__(self):
        info = {
            'debug': self.debug,
            'semaphore': self.semaphore,
            'url': HTTPClient.BASE
        }
        return make_repr(self, info)

    @classmethod
    def change_base(cls, base: str = None):
        """Change base for requests.

        Default base is ``http://www.boomlings.com/database/``,
        but it can be changed.

        Parameters
        ----------
        base: :class:`str`
            Base to change HTTPClient base to. If ``None`` or omitted, current base is preserved.
        """
        if base is None:
            return
        cls.BASE = base

    def set_default_semaphore(self, value: int = 250, *, loop=None):
        """Sets semaphore to :class:`asyncio.Semaphore` with given value and loop.

        Parameters
        ----------
        value: :class:`int`
            Value to set semaphore to. Default is ``200``.

        loop: :class:`asyncio.AbstractEventLoop`
            Event loop to pass to semaphore's constructor.
        """
        self.semaphore = asyncio.Semaphore(value, loop=loop)

    def set_semaphore(self, semaphore):
        """Sets semaphore of ``self`` to a given ``semaphore``.

        Parameters
        ----------
        semaphore: `Any`
            Semaphore to set. Preferably from ``asyncio`` module or subclasses of asyncio semaphores.
        """
        self.semaphore = semaphore

    def set_debug(self, debug: bool = False):
        """Set http client debugging.

        Parameters
        ----------
        debug: :class:`bool`
            Whether to set debug on or off.
        """
        self.debug = bool(debug)

    async def fetch(
        self, php: str, params: dict = None, get_cookies: bool = False,
        cookie: str = None, custom_base: str = None
    ):
        """|coro|

        Sends an HTTP Request to a Geometry Dash server and returns the response.

        .. note::
            This method does not raise any errors, but still,
            should be used carefully in programs.

        Parameters
        ----------
        php: :class:`str`
            The endpoint for the request URL, passed like this:

            .. code-block:: python3

                url = 'http://www.boomlings.com/database/' + php + '.php'

        params: Union[:class:`dict`, :class:`str`]
            Parameters to send with the request. Type ``dict`` is prefered.

        get_cookies: :class:`bool`
            Indicates whether to fetch a cookie. This is used
            in different account management actions.

        cookie: :class:`str`
            Cookie, represented as string. Technically should be
            catched with ``get_cookies`` set to ``True``.

        custom_base: :class:`str`
            Custom base using different Geometry Dash IP.
            By default ``http://boomlings.com/database/`` is used.

        Returns
        -------
        Union[:class:`bytes`, :class:`str`, :class:`int`, `None`]
            ``res`` with the following rules:

            If Exception occurs while sending request, returns ``None``.

            Returns an :class:`int`, representing server error code, e.g. ``-1``,
            if server responded with error.

            Otherwise, returns response from a server as a :class:`str` if successfully
            decodes it, and on fail returns :class:`bytes`.

            If a cookie is requested, returns a pair (``res``, ``c``) where c is a :class:`str` cookie.
        """
        base = HTTPClient.BASE if custom_base is None else custom_base
        url = base + php + '.php'

        method = 'GET' if params is None else 'POST'
        headers = None

        if cookie is not None:
            headers = {'Cookie': cookie}

        if self.debug:
            print('URL: {}, Data: {}'.format(url, params))

        async with aiohttp.ClientSession() as client:
            async with self.semaphore:
                try:
                    resp = await client.request(method, url, data=params, headers=headers)
                except HTTPClient.VALID_ERRORS:
                    return

                data = await resp.content.read()

                try:
                    res = data.decode()
                    if res.replace('-', '').isdigit():  # support for negative integers
                        return int(res)

                except UnicodeDecodeError:
                    res = data

                if self.debug:
                    print(res)

                if get_cookies:
                    c = str(resp.cookies).split(' ')[1]  # kinda tricky way
                    return res, c

                return res

    async def request(
        self, route: str, parameters: dict = None, custom_base: str = None,
        splitter: str = None, splitter_func: Callable[[str], list] = None,
        # 'error_codes' is a dict: {code: error_to_raise}
        error_codes: Dict[int, Exception] = None,
        # 'should_map': whether response should be mapped 'enum -> value' (dict)
        raise_errors: bool = True, should_map: bool = False,
        get_cookies: bool = False, cookie: str = None
    ):
        """|coro|

        A handy shortcut for fetching response from a server and formatting it.
        Basically does :meth:`HTTPClient.fetch` and operates on its result.

        Parameters
        ----------
        route: :class:`str`
            Same as ``php`` in :meth:`HTTPClient.fetch`.
        parameters: :class:`dict`
            Same as ``params`` in :meth:`HTTPClient.fetch`.
        custom_base: :class:`str`
            Same as ``custom_base`` in :meth:`HTTPClient.fetch`.
        splitter: :class:`str`
            A string to split the response with. If ``None``, splitting is passed.
        splitter_func: Callable[[:class:`str`], :class:`list`]
            A function that takes one argument of type string. Preferably should return a list.

            .. note::

                Either *splitter* can be specified, or *splitter_func*. Not both.

        error_codes: Dict[:class:`int`, :exc:`Exception`]
            A dictionary that response is checked against. ``Exception`` can be any Exception.
        raise_errors: :class:`bool`
            If ``False``, errors are not raised.
            (technically, just turns on ignoring ``error_codes``)
        should_map: :class:`bool`
            Whether should operate :meth:`.mapper_util.map` or not,
            considering that current response is converted to a list.
        get_cookies: :class:`bool`
            Same as ``get_cookies`` in :meth:`HTTPClient.fetch`
        cookie: :class:`str`
            Same as ``cookie`` in :meth:`HTTPClient.fetch`

        Raises
        ------
        :exc:`.HTTPNotConnected`
            GD Server has destroyed the connection, or machine has no connection.
            Raised when :meth:`HTTPClient.fetch` returns ``None`` and ``raise_errors`` is ``True``.

        :exc:`Exception`
            Exception specified in ``error_codes``, if present.

        Returns
        -------
        Union[:class:`int`, :class:`bytes`, :class:`str`, :class:`list`, :class:`dict`, `None`]
            If ``splitter``, ``splitter_func``, ``should_map`` and ``error_codes`` are omitted,
            return type is same as return of :meth:`HTTPClient.fetch` call.

            If ``error_codes`` is specified, and ``raise_errors`` is ``False``, returns ``None``
            when response is in ``error_codes``.

            If ``splitter`` is not ``None``, and ``should_map`` is
            ``False`` or omitted, returns :class:`list`.

            If ``splitter_func`` is not ``None`` and ``should_map`` is not present or ``False``,
            returns whatever passed function does return.

            If ``should_map`` is ``True``, returns :class:`dict`.
        """
        if parameters is None:
            parameters = {}

        if error_codes is None:
            error_codes = {}

        resp = await self.fetch(route, parameters, get_cookies, cookie, custom_base)

        if resp is None:
            if raise_errors:
                raise HTTPNotConnected()
            return

        if resp in error_codes:
            if raise_errors:
                raise error_codes.get(resp)
            return

        assert not (splitter is not None and splitter_func is not None)

        if not isinstance(resp, int):
            if splitter is not None:
                resp = resp.split(splitter)
            if splitter_func is not None:
                resp = splitter_func(resp)
            if should_map:
                resp = mapper_util.map(resp)
        return resp

    async def normal_request(self, url: str, data = None, method: str = None, **kwargs):
        """|coro|

        Same as doing :meth:`aiohttp.ClientSession.request`, where ``method`` is
        either given one or ``"GET"`` if ``data`` is None or omitted, and ``"POST"`` otherwise.
        """
        if method is None:
            method = 'GET' if data is None else 'POST'

        async with aiohttp.ClientSession() as client:
            try:
                resp = await client.request(method, url, data=data, **kwargs)

            except HTTPClient.VALID_ERRORS:
                raise HTTPNotConnected()

        return resp

http = HTTPClient()



"""Library that provides its users ability to interact with servers of Geometry Dash."""

__title__ = 'gd'
__author__ = 'NeKitDS'
__copyright__ = 'Copyright 2019 NeKitDS'
__license__ = 'MIT'
__version__ = '0.9.6'

from collections import namedtuple
import logging

from .abstractentity import AbstractEntity
from .abstractuser import AbstractUser
from .client import Client
from .colors import Colour, Color
from .colors import colors
from .comment import Comment
from .errors import *
from .friend_request import FriendRequest
from .iconset import IconSet
from .level import Level
from .level_packs import Gauntlet, MapPack
from .message import Message
from .session import GDSession
from .song import Song
from .unreguser import UnregisteredUser
from .user import UserStats, User
from .utils.captcha_solver import Captcha
from .utils.enums import *
from .utils.filters import Filters
from .utils.gdpaginator import paginate, Paginator
from .utils.http_request import http, HTTPClient
from .utils.params import Parameters
from .utils.save_parser import Save, SaveParser
from .utils.crypto.coders import Coder
from .utils.crypto.xor_cipher import XORCipher

from ._jokes import jokes  # why not?...

from .utils._async import enable_run_method as synchronise
synchronize = synchronise

from . import utils  # since asyncio.run() was introduced in 3.7, we have utils.run() function.


log = logging.getLogger(__name__)


VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')

def make_version_details(ver: str):
    cases = {
        'a' in ver: ('a', 'alpha'),
        'b' in ver: ('b', 'beta'),
        'rc' in ver: ('rc', 'candidate'),
        'f' in ver: ('f', 'final')
    }

    # we should not care about the separator if the releaselevel is 'final'
    splitter, releaselevel = cases.get(True, (' ', 'final'))
    splitted = ver.split(splitter)

    serial = 0
    if len(splitted) > 1:  # if there's a serial
        _, s = splitted
        serial = serial if not s else int(s)

    major, minor, micro = map(int, splitted[0].split('.'))

    return VersionInfo(
        major=major, minor=minor, micro=micro, releaselevel=releaselevel, serial=serial
    )

version_info = make_version_details(__version__)


def setup_logging(level: int = logging.DEBUG, *, stream=None, file=None, formatter=None):
    """Function that sets up logs of the module.

    Parameters
    ----------
    level: :class:`int`
        Level to set.

    stream: `Any`
        Stream to log messages into. ``stderr`` by default.

    file: `Any`
        File to log messages to. If not given, messages are logged into the ``stream``.

    formatter: :class:`str`
        Formatter to use. ``[LEVEL] (time) {gd.module}: Message`` by default.
    """
    handler = (
        logging.StreamHandler(stream) if file is None else logging.FileHandler(file)
    )

    if formatter is None:
        formatter = '[%(levelname)s] (%(asctime)s) {%(name)s}: %(message)s'

    handler.setFormatter(
        logging.Formatter(formatter)
    )

    log.addHandler(handler)

    log.setLevel(level)

# setting up the NullHandler here
try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

log.addHandler(NullHandler())


# delete not required stuff
del NullHandler, namedtuple



import asyncio

from ._async import *
from .enums import value_to_enum
from .search_utils import find, get
from .wrap_tools import *

def convert_to_type(obj: object, try_type: type, on_fail_type: type = None):
    """A function that tries to convert the given object to a provided type

    Parameters
    ----------
    obj: :class:`object`
        Any object to convert into given type.

    try_type: :class:`type`
        Type to convert an object to.

    on_fail_type: :class:`type`
        Type to convert an object on fail.
        If ``None`` or omitted, returns an ``obj``.
        On fail returns ``obj`` as well.

    Returns
    -------
    `Any`
        Object of given ``try_type``, on fail of type ``on_fail_type``, and
        ``obj`` if ``on_fail_type`` is ``None`` or failed to convert.
    """
    try:
        return try_type(obj)
    except Exception:  # failed to convert
        try:
            return on_fail_type(obj)
        except Exception:
            return obj



from .abstractentity import AbstractEntity
from .utils.wrap_tools import make_repr

class UnregisteredUser(AbstractEntity):
    """Class that represents Unregistered Users in Geometry Dash.
    This class is derived from :class:`.AbstractEntity`.
    """
    def __init__(self, id: int = 0, name: str = 'UnregisteredUser'):
        super().__init__(id=id)
        self._name = name

    def __repr__(self):
        info = {'id': self.id, 'name': self.name}
        return make_repr(self, info)

    @property
    def name(self):
        """:class:`str`: A name of the user. ``'UnregisteredUser'`` if not present."""
        return self._name



import asyncio

import logging
from typing import Union, Sequence

from .session import _session

from .utils.enums import value_to_enum, LeaderboardStrategy
from .utils.filters import Filters
from .utils.save_parser import Save
from .utils.wrap_tools import make_repr, check

from .utils.crypto.coders import Coder

from . import utils

log = logging.getLogger(__name__)


class Client:
    """A main class in the gd.py library, used for interacting with the servers of Geometry Dash.

    Parameters
    ----------
    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        The :class:`asyncio.AbstractEventLoop` to use for asynchronous operations.
        Defaults to ``None``, in which case the default event loop is used
        via :func:`asyncio.get_event_loop()`, meaning that
        :exc:`RuntimeError` can be raised if there is no current loop.

    Attributes
    ----------
    account_id: :class:`int`
        Account ID of the client. If not logged in, defaults to ``0``.
    id: :class:`int`
        ID (Player ID) of the client. If not logged in, defaults to ``0``.
    name: :class:`str`
        Name of the client. If not logged in, default is ``None``.
    password: :class:`str`
        Password of the client. ``None`` if not logged in.
    encodedpass: :class:`str`
        Encoded Password of the client. ``None`` on init as well.
    raw_save: :class:`str`
        Raw decoded save. If not loaded, defaults to ``''``.
    save: :class:`.Save`
        This is a namedtuple with format ``(completed, followed)``.
        Contains empty lists if not loaded.
    """
    def __init__(self, *, loop=None):
        self.session = _session
        self.loop = loop or asyncio.get_event_loop()
        self._set_to_defaults()

    def __repr__(self):
        info = {
            'is_logged': self.is_logged(),
            'account_id': self.account_id,
            'id': self.id,
            'name': self.name,
            'password': self.password
        }
        return make_repr(self, info)

    def _set_to_defaults(self):
        self.save = Save(completed=[], followed=[])
        self.raw_save = str()
        self.account_id = 0
        self.id = 0
        self.name = None
        self.password = None
        self.encodedpass = None

    def _upd(self, attr, value):
        setattr(self, attr, value)
        # update encodedpass if password was updated
        if attr == 'password':
            self.encodedpass = Coder.encode(type='accountpass', string=self.password)

    def is_logged(self):
        return (self.name is not None) and (self.password is not None)

    async def ping_server(self):
        """|coro|

        Pings ``boomlings.com/database`` and prints the time taken.
        Returns the ping as well.

        Returns
        -------
        :class:`float`
            Server ping, in milliseconds.
        """
        duration = await self.session.ping_server('http://boomlings.com/database/')

        print('GD Server ping: {}ms'.format(duration))

        return duration

    async def get_song(self, song_id: int = 0):
        """|coro|

        Fetches a song from Geometry Dash server.

        Parameters
        ----------
        song_id: :class:`int`
            An ID of the song to fetch.

        Raises
        ------
        :exc:`.MissingAccess`
            Song under given ID was not found or does not exist.

        :exc:`.SongRestrictedForUsage`
            Song was not allowed to use.

        Returns
        -------
        :class:`.Song`
            The song from the ID.
        """
        return await self.session.get_song(song_id)

    async def get_ng_song(self, song_id: int = 0):
        """|coro|

        Fetches a song from Newgrounds.

        This function is in most cases faster than :meth:`.Client.get_song`,
        and it does not raise errors if a song is banned on GD Server.

        Parameters
        ----------
        song_id: :class:`int`
            An ID of the song to fetch.

        Raises
        ------
        :exc:`.MissingAccess`
            Requested song is deleted from Newgrounds or does not exist.

        Returns
        -------
        :class:`.Song`
            The song found under given ID.
        """
        return await self.session.get_ng_song(song_id)

    async def get_user(self, account_id: int = 0):
        """|coro|

        Gets a user from Geometry Dash server.

        Parameters
        ----------
        account_id: :class:`int`
            An account ID of the user to fetch.

            .. note::

                If the given ID is equal to -1, a :class:`.UnregisteredUser` will be returned.

        Raises
        ------
        :exc:`.MissingAccess`
            User with given account ID was not found.

        Returns
        -------
        :class:`.User`
            The user from the ID. (if ID != -1)
        """
        return await self.session.get_user(account_id, return_only_stats=False, client=self)

    async def fetch_user(self, account_id: int = 0, *, stats: bool = False):
        """|coro|

        This is almost like :meth:`.Client.get_user`, except that it returns
        either :class:`.UserStats` or :class:`.AbstractUser` object.

        Parameters
        ----------
        account_id: :class:`int`
            An account ID of the user to fetch stats of.

            .. note::

                If the given ID is equal to -1, a :class:`.UnregisteredUser` will be returned.

        stats: :class:`bool`
            Whether to return :class:`.UserStats` or :class:`.AbstractUser`.
            By default returns :class:`.AbstractUser`.

        Raises
        ------
        :exc:`.MissingAccess`
            User with given account ID was not found, so fetching stats failed or user can not be returned.

        Returns
        -------
        Union[:class:`.UserStats`, :class:`.AbstractUser`]
            Abstract user or User stats from the ID. (if ID != -1)
        """
        user_stats = await self.session.get_user(account_id, return_only_stats=True, client=self)
        # return UserStats if needed, and AbstractUser otherwise.
        return user_stats if stats else user_stats.as_user()

    async def search_user(self, query: Union[int, str] = None):
        """|coro|

        Searches for a user on Geometry Dash servers.

        Parameters
        ----------
        query: Union[:class:`int`, :class:`str`]
            A query to search for user with. Either Player ID or Name.

        Raises
        ------
        :exc:`.MissingAccess`
            Nothing was found.

        Returns
        -------
        Union[:class:`.User`, :class:`.UnregisteredUser`]
            A User found when searching with the query.
        """
        return await self.session.search_user(query, return_abstract=False, client=self)

    async def find_user(self, query: Union[int, str] = None):
        """|coro|

        Fetches a user on Geometry Dash servers by given query.

        Works almost like :meth:`.Client.search_user`, except the fact that
        it returns :class:`.AbstractUser` or :class:`.UnregisteredUser`.

        Parameters
        ----------
        query: Union[:class:`int`, :class:`str`]
            A query to search for user with.

        Raises
        ------
        :exc:`.MissingAccess`
            No user was found.

        Returns
        -------
        Union[:class:`.AbstractUser`, :class:`.UnregisteredUser`]
            An AbstractUser or UnregisteredUser corresponding to the query.
        """
        return await self.session.search_user(query, return_abstract=True, client=self)

    async def get_daily(self):
        """|coro|

        Gets current daily level.

        Raises
        ------
        :exc:`.MissingAccess`
            Nothing found or invalid response was recieved.
        Returns
        -------
        :class:`.Level`
            Current daily level.
        """
        return await self.session.get_timely('daily', client=self)

    async def get_weekly(self):
        """|coro|

        Gets current weekly demon.

        Raises
        ------
        :exc:`.MissingAccess`
            Nothing found or invalid response was recieved.
        Returns
        -------
        :class:`.Level`
            Current weekly demon.
        """
        return await self.session.get_timely('weekly', client=self)

    async def get_level(self, level_id: int = 0):
        """|coro|

        Fetches a level from Geometry Dash servers.

        Parameters
        ----------
        level_id: :class:`int`
            An ID of the level to fetch.

            .. note::

                If the given ID is *n*, and ``0 > n >= -2`` is ``True``,
                this function will search for daily/weekly levels, however,
                it is not recommended to use since it can cause confusion.
                Use :meth:`.Client.get_daily` and :meth:`.Client.get_weekly` instead.

        Raises
        ------
        :exc:`.MissingAccess`
            Level with given ID was not found.

        Returns
        -------
        :class:`.Level`
            The level corresponding to given id.
        """
        return await self.session.get_level(level_id, client=self)

    async def get_gauntlets(self):
        """|coro|

        Fetches *The Lost Gauntlets*.

        Returns
        -------
        List[:class:`.Gauntlet`]
            All gauntlets retrieved, as list.
        """
        return await self.session.get_gauntlets(client=self)

    async def get_page_map_packs(self, page: int = 0, *, raise_errors: bool = True):
        """|coro|

        Fetches map packs on given page.

        Parameters
        ----------
        page: :class:`int`
            Page to look for map packs on.

        raise_errors: :class:`bool`
            Indicates whether :exc:`.NothingFound` should be raised. ``True`` by default.

        Returns
        -------
        List[:class:`.MapPack`]
            List of map packs retrieved.

        Raises
        ------
        :exc:`.NothingFound`
            No map packs were found at the given page.
        """
        return await self.session.get_page_map_packs(page=page, client=self)

    async def get_map_packs(self, pages: Sequence[int] = None):
        """|coro|

        Gets map packs on given ``pages``.

        Parameters
        ----------
        pages: Sequence[:class:`int`]
            Pages to search map packs on.

        Returns
        -------
        List[:class:`.MapPack`]
            List of map packs found.
        """
        if pages is None:
            pages = range(10)

        return await self.session.get_map_packs(pages=pages, client=self)

    async def test_captcha(self):
        """|coro|

        Tests Captcha solving, and prints the result.

        Returns
        -------
        :class:`int`
            The code of the Captcha.
        """
        code = await self.session.test_captcha()

        print(code)

        return code

    async def login(self, user: str, password: str):
        """|coro|

        Tries to login with given parameters.

        Parameters
        ----------
        user: :class:`str`
            A username of the account to log into.

        password: :class:`str`
            A password of the account to log into.

        Raises
        ------
        :exc:`.LoginFailure`
            Given account credentials are not correct.
        """
        await self.session.login(client=self, user=user, password=password)
        log.info("Logged in as %r, with password %r.", user, password)

    @check.is_logged()
    async def get_page_levels(self, page: int = 0, *, raise_errors: bool = True):
        """|coro|

        Gets levels of a client from a server.

        .. note::

            This method requires client to be logged in.

        Parameters
        ----------
        page: :class:`int`
            Page to look levels at.

        raise_errors: :class:`bool`
            Indicates whether :exc:`.MissingAccess` should be raised. ``True`` by default.

        Returns
        -------
        List[:class:`.Level`]
            All levels found, as list. Might be an empty list.

        Raises
        ------
        :exc:`.MissingAccess`
            No levels were found.
        """
        filters = Filters.setup_by_user()
        return await self.search_levels_on_page(filters=filters, raise_errors=raise_errors)

    @check.is_logged()
    async def get_levels(self, pages: Sequence[int] = None):
        """|coro|

        Searches for levels on given pages.

        .. note::

            This method requires authorised client.

        Parameters
        ----------
        pages: Sequence[:class:`int`]
            Pages to look for levels at.

        Returns
        -------
        List[:class:`.Level`]
            All levels found, as list, which might be empty.
        """
        filters = Filters.setup_by_user()
        return await self.search_levels(pages=pages, filters=filters)

    @check.is_logged()
    async def load(self):
        """|coro|

        Loads save from a server and parses it.
        Sets :attr:`.Client.save` to :class:`.Save` namedtuple ``(completed, followed)``.
        """
        success = await self.session.load_save(client=self)

        if success:
            log.info('Successfully loaded a save.')
        else:
            log.warning('Failed to load a save.')

    def close(self, message: str = 'Logged out, no additional description.'):
        """*Closes* client.

        Basically sets its password and username to ``None``, which
        actually implies that client logs out.

        Parameters
        ----------
        message: :class:`str`
            A message to print after closing.
        """
        self._set_to_defaults()

        print(message)

        log.info('Has logged out with message: %r', message)

    @check.is_logged()
    async def get_friends(self):
        """|coro|

        Get all friends of a client.

        Returns
        -------
        List[:class:`.AbstractUser`]
            All friends retrieved, as list.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to fetch friends of a client.

        :exc:`.NothingFound`
            No friends were found. Sadly...
        """
        return await self.session.get_friends(client=self)

    @check.is_logged()
    async def to_user(self):
        """|coro|

        Gets user with :attr:`.Client.account_id`,
        which means that client should be logged in.

        Returns
        -------
        :class:`.User`
            User corresponding to :attr:`.Client.account_id`.
        """
        return await self.get_user(self.account_id)

    @check.is_logged()
    async def get_page_messages(
        self, sent_or_inbox: str = 'inbox', page: int = 0, *, raise_errors: bool = True
    ):
        """|coro|

        Gets messages on a specified page.

        Requires logged in client.

        Parameters
        ----------
        sent_or_inbox: :class:`str`
            The type of messages to look for. Either *inbox* or *sent*.

        page: :class:`int`
            Number of page to look at.

        raise_errors: :class:`bool`
            Indicates whether errors should be raised.

        Returns
        -------
        List[:class:`.Message`]
            List of messages found. Can be empty.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to get messages. Raised if ``raise_errors`` is ``True``.

        :exc:`.NothingFound`
            No messages were found. Raised if ``raise_errors`` is ``True``.
        """
        return await self.session.get_page_messages(
            sent_or_inbox=sent_or_inbox, page=page,
            raise_errors=raise_errors, client=self
        )

    @check.is_logged()
    async def get_messages(
        self, sent_or_inbox: str = 'inbox', pages: Sequence[int] = None
    ):
        """|coro|

        Retrieves messages from given ``pages``.

        Parameters
        ----------
        sent_or_inbox: :class:`str`
            Type of messages to retrieve. Either `'sent'` or `'inbox'`.
            Defaults to the latter.

        pages: Sequence[:class:`int`]
            Pages to look at, represented as a finite sequence, so iterations can be performed.

        Returns
        -------
        List[:class:`.Message`]
            List of messages found. Can be an empty list.
        """
        if pages is None:
            pages = range(10)

        return await self.session.get_messages(
            sent_or_inbox=sent_or_inbox, pages=pages, client=self
        )

    @check.is_logged()
    async def get_page_friend_requests(
        self, sent_or_inbox: str = 'inbox', page: int = 0, *, raise_errors: bool = True
    ):
        """|coro|

        Gets friend requests on a specified page.

        Requires logged in client.

        Parameters
        ----------
        sent_or_inbox: :class:`str`
            The type of friend requests to look for. Either *inbox* or *sent*.

        page: :class:`int`
            Number of page to look at.

        raise_errors: :class:`bool`
            Indicates whether errors should be raised.

        Returns
        -------
        List[:class:`.FriendRequest`]
            List of friend requests found. Can be empty.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to get friend requests. Raised if ``raise_errors`` is ``True``.

        :exc:`.NothingFound`
            No friend requests were found. Raised if ``raise_errors`` is ``True``.
        """
        return await self.session.get_page_friend_requests(
            sent_or_inbox=sent_or_inbox, page=page,
            raise_errors=raise_errors, client=self
        )

    @check.is_logged()
    async def get_friend_requests(
        self, sent_or_inbox: str = 'inbox', pages: Sequence[int] = None
    ):
        """|coro|

        Retrieves friend requests from given ``pages``.

        Parameters
        ----------
        sent_or_inbox: :class:`str`
            Type of friend requests to retrieve. Either `'sent'` or `'inbox'`.
            Defaults to the latter.

        pages: Sequence[:class:`int`]
            Pages to look at, represented as a finite sequence, so iterations can be performed.

        Returns
        -------
        List[:class:`.FriendRequests`]
            List of friend requests found. Can be an empty list.
        """
        if pages is None:
            pages = range(10)

        return await self.session.get_friend_requests(
            sent_or_inbox=sent_or_inbox, pages=pages, client=self
        )

    @check.is_logged()
    def get_parse_dict(self):
        return {k: getattr(self, k) for k in ('name', 'id', 'account_id')}

    def as_user(self):
        return self.session.to_user(self.get_parse_dict(), client=self)

    async def get_top(self, strategy: Union[int, str, LeaderboardStrategy] = 0, *, count: int = 100):
        """|coro|

        Fetches user top by given strategy.

        Example:

        .. code-block:: python3

            # getting top 10 creators
            top10_creators = await client.get_top('creators', count=10)

        .. note::

            Players Top 100 has stopped refreshing in 2.1 version of Geometry Dash.
            However, you can fetch it by searching using ``'relative'`` strategy
            and giving huge ``count`` argument.

            Also, please note that searching with ``'friends'`` and ``'relative'`` strategies
            requires logged in client.

        Parameters
        ----------
        strategy: Union[:class:`int`, :class:`str`, :class:`.LeaderboardStrategy`]
            Strategy to apply when searching.


        """
        strategy = value_to_enum(LeaderboardStrategy, strategy)
        return await self.session.get_top(strategy=strategy, count=count, client=self)

    async def get_leaderboard(
        self, strategy: Union[int, str, LeaderboardStrategy] = 0, *, count: int = 100):
        """|coro|

        This is an alias for :meth:`.Client.get_top`.
        """
        return await self.get_top(strategy, count=count)

    @check.is_logged()
    async def get_account_url(self):
        """|coro|

        Fetches account URL. It is used in Save retrieving.

        Returns
        -------
        :class:`str`
            Link of format ``http://<ip>/database/``.
        """
        return await self.session.get_account_url(self)

    @check.is_logged()
    async def post_comment(self, content: str):
        """|coro|

        Post a profile comment.

        Parameters
        ----------
        content: :class:`str`
            The content of a comment.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to post a comment.
        """
        await self.session.post_comment(content, client=self)

        log.debug("Posted a comment. Content: %s", content)

    @check.is_logged()
    async def edit(self, name: str = None, password: str = None):
        """|coro|

        Tries to edit credentials of a client, if logged in.

        .. note::

            See :meth:`.Client.test_captcha` to see information about how to treat this function.

        Parameters
        ----------
        name: :class:`str`
            A name to change logged account's username to. Defaults to ``None``.

        password: :class:`str`
            A password to change logged account's username to. Defaults to ``None``.

        Raises
        ------
        :exc:`.FailedToChange`
            Failed to change the credentials.
        """
        await self.session.edit(name=name, password=password, client=self)

        # if no errors occured, log changes.
        if name is not None:
            log.debug('Changed username to: %s', name)

        if password is not None:
            log.debug('Changed password to: %s', password)


    @check.is_logged()
    async def update_profile(
        self, stars: int = 0, demons: int = 0, diamonds: int = 0, main_icon: int = 1,
        main_icon_type: int = 0, color_1: int = 0, color_2: int = 3, coins: int = 0,
        user_coins: int = 0, cube: int = 1, ship: int = 1, ball: int = 1, ufo: int = 1,
        wave: int = 1, robot: int = 1, spider: int = 1, explosion: int = 1,
        has_glow: bool = False, like_user=None
    ):
        pass

    @check.is_logged()
    async def update_settings(
        self, *, msg: int = None, friend_req: int = None, comments: int = None,
        youtube: str = None, twitter: str = None, twitch: str = None
    ):
        """|coro|

        Updates profile settings of a client.

        .. note::

            For parameter in parameters, if parameter is ``None`` or omitted,
            it will be set to the current policy/link of the user corresponding
            to this client; that implies that running the following:

            .. code-block:: python3

                await client.update_settings()

            will cause no effect on profile settings.

        Parameters
        ----------
        msg: Union[:class:`int`, :class:`.MessagePolicyType`]
            New message policy.
        friend_req: Union[:class:`int`, :class:`.FriendRequestPolicyType`]
            New friend request policy.
        comments: Union[:class:`int`, :class:`.CommentPolicyType`]
            New comment history policy.
        youtube: :class:`str`
            New youtube channel string. (not link)
        twitter: :class:`str`
            New twitter profile name.
        twitch: :class:`str`
            New twitch profile name.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to update profile.
        """
        profile_dict = {
            'msg_policy': msg,
            'friend_req_policy': friend_req,
            'comments_policy': comments,
            'youtube': youtube,
            'twitter': twitter,
            'twitch': twitch
        }

        self_user = await self.to_user()

        args = []

        for attr, value in profile_dict.items():
            tmp = getattr(self_user, attr) if value is None else value
            if tmp is None:
                s = ''
            else:
                s = utils.convert_to_type(tmp, int, str)

            args.append(s)

        await self.session.update_settings(*args, client=self)

    async def search_levels_on_page(
        self, page: int = 0, query: str = '', filters: Filters = None, user=None,
        *, raise_errors: bool = True
    ):
        """|coro|

        Searches levels on given page by given query, applying filters as well.

        Parameters
        ----------
        page: :class:`int`
            A page to search levels on.

        query: :class:`int`
            A query to search with.

        filters: :class:`.Filters`
            Filters to apply, as an object.

        user: Union[:class:`int`, :class:`.AbstractUser`, :class:`.User`]
            A user to search levels by. (if :class:`.Filters` has parameter ``strategy``
            equal to :class:`.SearchStrategy` ``BY_USER``. Can be omitted, then
            logged in client is required.)

        raise_errors: :class:`bool`
            Whether or not to raise errors.

        Returns
        -------
        List[:class:`.Level`]
            Levels found on given page.

        Raises
        ------
        ``None`` [yet]
        """
        return await self.session.search_levels_on_page(
            page=page, query=query, filters=filters, user=user, raise_errors=raise_errors, client=self
        )

    async def search_levels(
        self, query: str = '', filters: Filters = None, user=None,
        pages: Sequence[int] = None
    ):
        """|coro|

        Searches levels on given pages.

        Parameters
        ----------
        query: :class:`int`
            A query to search with.

        filters: :class:`.Filters`
            Filters to apply, as an object.

        user: Union[:class:`int`, :class:`.AbstractUser`, :class:`.User`]
            A user to search levels by. (if :class:`.Filters` has parameter ``strategy``
            equal to :class:`.SearchStrategy` ``BY_USER``. Can be omitted, then
            logged in client is required.)

        pages: Sequence[:class:`int`]
            Pages to look at, represented as a finite sequence, so iterations can be performed.

        Returns
        -------
        List[:class:`.Level`]
            List of levels found. Can be an empty list.
        """
        if pages is None:
            pages = range(10)

        return await self.session.search_levels(
            query=query, filters=filters, user=user, pages=pages, client=self
        )

    def run(self, coro, *, debug: bool = False):
        """A handy shortcut for :func:`.utils.run`.

        This is equivalent to:

        .. code-block:: python3

            gd.utils.run(coro, loop=self.loop, debug=debug)
        """
        return utils.run(coro, loop=self.loop, debug=debug)

    def event(self, coro):
        """A decorator that registers an event to listen to.

        Events must be a |coroutine_link|_, if not, :exc:`TypeError` is raised.

        Example
        -------

        .. code-block:: python3

            @client.event
            async def on_level_rated(level):
                print(level.name)

        Raises
        ------
        TypeError
            The coroutine passed is not actually a coroutine.
        """

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("event registered must be a coroutine function.")

        setattr(self, coro.__name__, coro)
        log.debug("%s has been successfully registered as an event.", coro.__name__)

        return coro



from typing import Any, Sequence

from .enums import (
    NEnum, LevelDifficulty, DemonDifficulty,
    SearchStrategy, LevelLength, value_to_enum
)
from .wrap_tools import make_repr

class Filters:
    """Class that adds ability to create custom filters to apply when searching for levels.

    Attributes
    ----------
    strategy: Optional[Union[:class:`int`, :class:`str`, :class:`SearchStrategy`]]
        A strategy to apply.
    difficulty: Optional[Sequence[Union[:class:`int`, :class:`str`, :class:`LevelDifficulty`]]]
        Level difficulties to apply. Either sequence of many or just one difficulty.
    demon_difficulty: Optional[Union[:class:`int`, :class:`str`, :class:`DemonDifficulty`]]
        **One** Demon difficulty to apply as a filter.
    length: Sequence[Union[:class:`int`, :class:`str`, :class:`LevelLength`]]
        Same as ``difficulty``, but for lengths.
    uncompleted: Optional[:class:`bool`]
        Whether to search for uncompleted levels only. If true, ``completed_levels`` is required.
    only_completed: Optional[:class:`bool`]
        Opposite of ``uncompleted``.
    completed_levels: Optional[Sequence[Union[:class:`int`, :class:`.Level`]]]
        A sequence of completed levels.
    require_coins: Optional[:class:`bool`]
        Whether levels must have coins in them.
    featured: Optional[:class:`bool`]
        Whether levels **must** be featured.
    epic: Optional[:class:`bool`]
        Whether levels ought to be epic.
    rated: Optional[:class:`bool`]
        This defaults to ``None``. If true, will search for only rated levels,
        if false, will search for only not rated levels. If ``None``, makes no changes.
    require_two_player: Optional[:class:`bool`]
        Whether a level must be for two players.
    song_id: Optional[:class:`int`]
        An ID of the song that should be used.
    use_custom_song: Optional[:class:`int`]
        Whether a custom song should be used. Requires ``song_id`` to be defined.
    followed: Optional[Sequence[Union[:class:`int`, :class:`.AbstractUser`, :class:`.User`]]]
        Followed users, basically this will work as ``BY_USER`` ``strategy``, but for many users.
    require_original: Optional[:class:`bool`]
        Whether levels must be original.
    """
    def __init__(
        self, strategy=0, difficulty=None, demon_difficulty=None, length=None,
        uncompleted=False, only_completed=False, completed_levels=[],
        require_coins=False, featured=False, epic=False, rated=None,
        require_two_player=False, song_id=None, use_custom_song=False,
        followed=[], require_original=False
    ):
        if isinstance(difficulty, (int, str, NEnum)):
            difficulty = [difficulty]

        if isinstance(length, (int, str, NEnum)):
            length = [length]

        self.strategy = value_to_enum(SearchStrategy, strategy)

        ss = (self.strategy in (5, 6, 16))

        self.difficulty = None if ((difficulty is None) or ss) else tuple(
            map(lambda diff: value_to_enum(LevelDifficulty, diff), difficulty)
        )

        self.length = None if ((length is None) or ss) else tuple(
            map(lambda l_length: value_to_enum(LevelLength, l_length), length)
        )

        self.demon_difficulty = (
            None if (
                (demon_difficulty is None) or ss
            ) else value_to_enum(DemonDifficulty, demon_difficulty)
        )

        if self.demon_difficulty is not None:
            self.diffculty = LevelDifficulty.DEMON

        self.uncompleted = uncompleted if completed_levels else False
        self.only_completed = only_completed if completed_levels else False
        self.completed_levels = list(completed_levels)
        self.require_coins = require_coins
        self.require_two_player = require_two_player
        self.rated = rated
        self.featured = featured
        self.epic = epic
        self.song_id = song_id
        self.use_custom_song = use_custom_song
        self.followed = list(followed)
        self.require_original = require_original
        self.set = (
            'strategy', 'difficulty', 'demon_difficulty', 'length',
            'uncompleted', 'only_completed', 'completed_levels',
            'require_coins', 'require_two_player', 'rated',
            'featured', 'epic', 'song_id', 'use_custom_song',
            'followed', 'require_original'
        )

    def __repr__(self):
        info = {k: repr(getattr(self, k)) for k in self.set}
        return make_repr(self, info)

    @classmethod
    def setup_empty(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    @classmethod
    def setup_simple(cls, *args, **kwargs):
        return cls(strategy=2)

    @classmethod
    def setup_by_user(cls, *args, **kwargs):
        return cls(strategy=5, *args, **kwargs)

    @classmethod
    def setup_with_song(cls, song_id: int, is_custom: bool = True, *args, **kwargs):
        return cls(strategy=2, song_id=song_id, use_custom_song=is_custom, *args, **kwargs)

    @classmethod
    def setup_level_pack(cls):
        return cls(strategy=10)

    @classmethod
    def setup_with_followed(cls, followed: Sequence[int], *args, **kwargs):
        return cls(strategy=12, followed=followed, *args, **kwargs)

    @classmethod
    def setup_by_friends(cls, *args, **kwargs):
        return cls(strategy=13, *args, **kwargs)

    @classmethod
    def setup_client_followed(cls, client, *args, **kwargs):
        return cls(strategy=12, followed=client.save.followed, *args, **kwargs)

    @classmethod
    def setup_client_completed(cls, client, *args, **kwargs):
        return cls(strategy=2, completed_levels=client.save.completed, *args, **kwargs)

    def to_parameters(self):
        main = {
            'type': self.strategy.value,
            'diff': '-' if self.difficulty is None else _join(',', self.difficulty),
            'len': '-' if self.length is None else _join(',', self.length),
            'uncompleted': int(self.uncompleted),
            'onlyCompleted': int(self.only_completed),
            'featured': int(self.featured),
            'original': int(self.require_original),
            'twoPlayer': int(self.require_two_player),
            'coins': int(self.require_coins),
            'epic': int(self.epic)
        }

        if self.demon_difficulty is not None:
            main.update(
                {'demonFilter': self.demon_difficulty.value}
            )

        if self.uncompleted or self.only_completed:  # adds only if completed_levels not empty
            main.update(
                {'completedLevels': _join(',', self.completed_levels, wrap_with='({})')}
            )

        if self.rated is not None:
            main.update(
                {('star' if self.rated else 'noStar'): 1}
            )

        if self.strategy == SearchStrategy.FOLLOWED:
            main.update(
                {'followed': _join(',', self.followed)}
            )

        if self.song_id is not None:
            main.update({
                'song': int(self.song_id),
                'isCustom': int(self.use_custom_song)
            })

        filters = {k: str(v) for k, v in main.items()}
        return filters

def _join(string: str, elements: Sequence[Any], *, wrap_with: str = '{}'):

    def func(element):
        to_str = element
        if isinstance(element, NEnum):  # Enum
            to_str = element.value
        elif hasattr(element, 'account_id'):  # AbstractUser
            to_str = element.account_id
        elif hasattr(element, 'id'):  # Level
            to_str = element.id
        return str(to_str)

    return wrap_with.format(string.join(map(func, elements)))



from .abstractentity import AbstractEntity
from .utils.filters import Filters
from .utils.wrap_tools import make_repr

class Gauntlet(AbstractEntity):
    """Class that represents *The Lost Gauntlet* in Geometry Dash.
    This class is derived from :class:`.AbstractEntity`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self.options = options
        self._levels = ()

    def __repr__(self):
        info = {
            'id': self.id,
            'name': self.name,
            'levels': self.level_ids
        }

        if self.levels:
            info.update({'levels': self.levels})

        return make_repr(self, info)

    @property
    def name(self):
        """:class:`str`: Name of the Gauntlet."""
        return self.options.get('name')

    @property
    def level_ids(self):
        """Tuple[:class:`int`]: Tuple of level IDs that are contained in the gauntlet."""
        return self.options.get('level_ids')

    @property
    def levels(self):
        """Tuple[:class:`.Level`]: Tuple containing levels of the Gauntlet.

        Can be retrieved with :meth:`.Gauntlet.get_levels`.
        """
        return self._levels

    async def get_levels(self):
        """|coro|

        Retrieves levels of a Gauntlet or its subclass.

        Returns
        -------
        List[:class:`.Level`]
            List of levels that are found.
        """
        filters, query = Filters.setup_level_pack(), ','.join(map(str, self.level_ids))

        levels = await self._client.search_levels_on_page(query=query, filters=filters)

        self._levels = tuple(levels)
        return levels


class MapPack(Gauntlet):
    """Class that represents *Map Pack* in Geometry Dash.
    This class is derived from :class:`.Gauntlet`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self.options = options

    def __repr__(self):
        info = {
            'id': self.id,
            'name': self.name,
            'color': self.color,
            'levels': self.level_ids
        }

        if self.levels:
            info.update({'levels': self.levels})

        return make_repr(self, info)

    @property
    def stars(self):
        """:class:`int`: Amount of stars this map pack has."""
        return self.options.get('stars')

    @property
    def coins(self):
        """:class:`int`: Number of coins this map pack grants on completion."""
        return self.options.get('coins')

    @property
    def difficulty(self):
        """:class:`.LevelDifficulty`: Average difficulty of a map pack."""
        return self.options.get('difficulty')

    @property
    def color(self):
        """:class:`.Colour`: Colour of a map pack."""
        return self.options.get('color')



class Index:
    """Class that contains all indexes required for operating on GD Server responses."""
    # Indexes '10X' are custom made by this library, and have nothing to do with the servers.
    # All indexes for gd levels
    LEVEL_ID = 1
    LEVEL_NAME = 2
    LEVEL_DESCRIPTION = 3
    LEVEL_DATA = 4
    LEVEL_VERSION = 5
    LEVEL_CREATOR_ID = 6
    LEVEL_DIFFICULTY = 9
    LEVEL_DOWNLOADS = 10
    LEVEL_AUDIO_TRACK = 12
    LEVEL_GAME_VERSION = 13
    LEVEL_LIKES = 14
    LEVEL_LENGTH = 15
    LEVEL_IS_DEMON = 17
    LEVEL_STARS = 18
    LEVEL_FEATURED_SCORE = 19
    LEVEL_IS_AUTO = 25
    LEVEL_PASS = 27
    LEVEL_UPLOADED_TIMESTAMP = 28
    LEVEL_LAST_UPDATED_TIMESTAMP = 29
    LEVEL_ORIGINAL = 30
    LEVEL_SONG_ID = 35
    LEVEL_COIN_COUNT = 37
    LEVEL_COIN_VERIFIED = 38
    LEVEL_REQUESTED_STARS = 39
    LEVEL_IS_EPIC = 42
    LEVEL_DEMON_DIFFICULTY = 43
    LEVEL_OBJECT_COUNT = 45
    LEVEL_TIMELY_TYPE = 101
    LEVEL_TIMELY_INDEX = 102
    LEVEL_TIMELY_COOLDOWN = 103
    # all indexes for songs
    SONG_ID = 1
    SONG_TITLE = 2
    SONG_AUTHOR = 4
    SONG_SIZE = 5
    SONG_URL = 10
    # all indexes for users
    USER_NAME = 1
    USER_PLAYER_ID = 2
    USER_STARS = 3
    USER_DEMONS = 4
    USER_TOP_PLACE = 6
    USER_CREATOR_POINTS = 8
    USER_ICON = 9
    USER_COLOR_1 = 10
    USER_COLOR_2 = 11
    USER_SECRET_COINS = 13
    USER_ICON_TYPE = 14
    USER_GLOW_OUTLINE = 15
    USER_ACCOUNT_ID = 16
    USER_COINS = 17
    USER_PRIVATE_MESSAGE_POLICY = 18
    USER_FRIEND_REQUEST_POLICY = 19
    USER_YOUTUBE = 20
    USER_ICON_CUBE = 21
    USER_ICON_SHIP = 22
    USER_ICON_BALL = 23
    USER_ICON_UFO = 24
    USER_ICON_WAVE = 25
    USER_ICON_ROBOT = 26
    USER_GLOW_OUTLINE_2 = 28
    USER_GLOBAL_RANK = 30
    USER_ICON_SPIDER = 43
    USER_TWITTER = 44
    USER_TWITCH = 45
    USER_DIAMONDS = 46
    USER_DEATH_EFFECT = 47
    USER_ROLE = 49
    USER_COMMENT_HISTORY_POLICY = 50
    # all indexes for messages
    MESSAGE_ID = 1
    MESSAGE_SENDER_ACCOUNT_ID = 2
    MESSAGE_SENDER_ID = 3
    MESSAGE_SUBJECT = 4
    MESSAGE_BODY = 5
    MESSAGE_SENDER_NAME = 6
    MESSAGE_TIMESTAMP = 7
    MESSAGE_IS_READ = 8
    MESSAGE_INDICATOR = 9  # when == 1 - the message is sent TO MESSAGE_SENDER
    # all indexes for comments
    COMMENT_LEVEL_ID = 1
    COMMENT_BODY = 2
    COMMENT_AUTHOR_ID = 3
    COMMENT_RATING = 4
    COMMENT_ID = 6
    COMMENT_IS_SPAM = 7
    COMMENT_TIMESTAMP = 9
    COMMENT_LEVEL_PERCENTAGE = 10
    COMMENT_USER_STATUS = 11
    COMMENT_COLOR = 12
    COMMENT_TYPE = 101
    # all indexes for friend requests (at least all I need lol)
    REQUEST_SENDER_NAME = 1
    REQUEST_SENDER_ID = 2
    REQUEST_SENDER_ACCOUNT_ID = 16
    REQUEST_ID = 32
    REQUEST_BODY = 35
    REQUEST_TIMESTAMP = 37
    REQUEST_STATUS = 41
    REQUEST_INDICATOR = 101
    # all indexes for map packs
    MAP_PACK_ID = 1
    MAP_PACK_NAME = 2
    MAP_PACK_LEVEL_IDS = 3
    MAP_PACK_STARS = 4
    MAP_PACK_COINS = 5
    MAP_PACK_DIFFICULTY = 6
    MAP_PACK_COLOR = 7
    # all indexes for gauntlets
    GAUNTLET_ID = 1
    GAUNTLET_LEVEL_IDS = 3



"""Somewhat useful 'python -m gd' implementation"""

import argparse
import sys

import gd
import pkg_resources
import aiohttp
import PIL
import platform

con_v = '0.2.0'  # gd_console version

def show_version():
    entries = []

    entries.append('- Python v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}'.format(sys.version_info))

    version_info = gd.version_info
    entries.append('- gd.py v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}'.format(version_info))

    if version_info.releaselevel != 'final':
        pkg = pkg_resources.get_distribution('gd.py')
        if pkg:
            entries.append('    - gd.py pkg_resources: v{0}'.format(pkg.version))

    entries.append('- [gd_console] v{0}'.format(con_v))

    entries.append('- aiohttp v{0.__version__}'.format(aiohttp))

    entries.append('- Pillow (PIL) v{0.__version__}'.format(PIL))

    uname = platform.uname()
    entries.append('- System Info: {0.system} {0.release} {0.version}'.format(uname))

    print('\n'.join(entries))


def show_docs():
    _docs = 'https://gdpy.readthedocs.io/en/latest'

    print('- gd.py docs: [{}]'.format(_docs))

def main():
    # make parser
    parser = argparse.ArgumentParser(description='gd.py console commands', prog='gd')
    # add --version
    parser.add_argument('-v', '--version', help='show versions (gd.py, python, etc.)',
        action='store_true', default=False)
    # add --docs
    parser.add_argument('-d', '--docs', help='show links to gd.py docs',
        action='store_true', default=False)
    # parse args
    args = parser.parse_args()
    # run functions
    if args.version:
        show_version()
    if args.docs:
        show_docs()

# run main
if __name__ == '__main__':
    main()



from ..utils.wrap_tools import make_repr

class Rectangle:
    def __init__(self, x, y, w, h):
        self.x, self.y = x, y
        self.width, self.height = w, h

    @property
    def size(self):
        return self.width, self.height

    def get_coords(self):
        return self.x, self.y

class Sprite:
    def __init__(self, **options):
        self.options = options

    def __repr__(self):
        info = {
            'name': repr(self.name),
            'id': self.id,
            'size': 'x'.join(map(str, self.size)),
            'is_rotated': self.is_rotated()
        }
        return make_repr(self, info)

    @property
    def name(self):
        return self.options.get('name')

    @property
    def id(self):
        return self.options.get('id')

    @property
    def offset(self):
        return self.options.get('offset')

    @property
    def size(self):
        return self.options.get('size')

    @property
    def source_size(self):
        return self.options.get('source_size')

    def update_offset(self, x1, y1):
        x, y = self.offset
        x += x1
        y += y1
        self.options['offset'] = (x, y)

    def is_rotated(self):
        return self.options.get('is_rotated')

    def get_rectangle(self):
        return self.options.get('rectangle')

    def duplicate(self):
        return Sprite(
            name = self.name + 'D',
            id = self.id,
            offset = self.offset,
            size = self.size,
            source_size = self.source_size,
            is_rotated = self.is_rotated(),
            rectangle = self.get_rectangle()
        )


class mapper_util:
    def map(item):
        res = {}
        for i in range(0, len(item), 2):
            new = item[i+1]
            new = new if not _is_str_digit(new) else int(new)
            res[int(item[i])] = new
        return res

    def normalize(item):
        res = str(item).replace('-', '+').replace('_', '/')
        return _pad(res)

    def prepare_sending(item):
        res = str(item).replace('+', '-').replace('/', '_')
        return _pad(res)

def _pad(res: str):
    # pad a string to be divisible by 4
    ilen = len(res)/4

    if not ilen.is_integer():
        n = _ceil(ilen)*4 - len(res)

        res += '=' * n

    return res

def _ceil(n: float):
    x = round(n)
    return x+1 if x < n else x

def _is_str_digit(string: str):
    return string.replace('-', '').isdigit()



from setuptools import setup
import re

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('gd/__init__.py') as f:
    try:
        version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)
    except AttributeError:
        raise RuntimeError('Version is not set.') from None

readme = ''
with open('README.rst') as f:
    readme = f.read()

extras_require = {  # gotta work on that soon
    'docs': [
        'sphinx',
        'sphinxcontrib_trio',
        'sphinxcontrib-websupport'
    ]
}

setup(
    name='gd.py',
    author='NeKitDS',
    author_email='gdpy13@gmail.com',
    url='https://github.com/NeKitDS/gd.py',
    project_urls={
        "Documentation": "https://gdpy.readthedocs.io/en/latest",
        "Issue tracker": "https://github.com/NeKitDS/gd.py/issues",
    },
    version=version,
    packages=[
        'gd', 'gd.utils', 'gd.utils.crypto',
        'gd.graphics', 'gd.events', 'gd.api'
    ],
    license='MIT',
    description='A Geometry Dash API wrapper for Python',
    long_description=readme,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_require,
    python_requires='>=3.5.3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Natural Language :: English',
        'Operating System :: OS Independent'
    ],
    entry_points={
        'console_scripts': [
            'gd = gd.__main__:main',
        ]
    }
)



from typing import Sequence, Union

from .abstractentity import AbstractEntity
from .session import _session

from .utils.enums import CommentStrategy, value_to_enum
from .utils.filters import Filters
from .utils.wrap_tools import make_repr, check

class AbstractUser(AbstractEntity):
    """Class that represents an Abstract Geometry Dash User.
    This class is derived from :class:`.AbstractEntity`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self.options = options

    def __repr__(self):
        info = {
            'name': repr(self.name),
            'id': self.id,
            'account_id': self.account_id
        }
        return make_repr(self, info)

    @property
    def name(self):
        """:class:`str`: String representing name of the user."""
        return self.options.get('name')

    @property
    def account_id(self):
        """:class:`int`: Account ID of the user."""
        return self.options.get('account_id')

    @property
    def _dict_for_parse(self):
        return {
            k: getattr(self, k) for k in ('name', 'id', 'account_id')
        }

    def as_user(self):
        """Returns :class:`.AbstractUser` object.

        This is used mainly in subclasses.

        Returns
        -------
        :class:`.AbstractUser`
            Abstract User from given object.
        """
        return _session.to_user(self._dict_for_parse, client=self._client)

    async def to_user(self):
        """|coro|

        Convert ``self`` to :class:`.User` object.

        Returns
        -------
        :class:`.User`
            A user object corresponding to the abstract one.
        """
        return await self._client.get_user(self.account_id)

    async def update(self):
        """|coro|

        Update ``self``.
        """
        new = await self._client.fetch_user(self.account_id)
        self.options = new.options

    async def send(self, subject: str, body: str, *, from_client=None):
        """|coro|

        Send the message to ``self``. Requires logged client.

        Parameters
        ----------
        subject: :class:`str`
            The subject of the message.

        body: :class:`str`
            The body of the message.

        from_client: :class:`.Client`
            A logged in client to send a message from. If ``None``,
            defaults to a client attached to this user.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to send a message.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'send')
        await _session.send_message(target=self, subject=subject, body=body, client=client)

    async def block(self, *, from_client=None):
        """|coro|

        Block a user. Requires logged in client.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to block a user from. If ``None``,
            defaults to a client attached to this user.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to block a user.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'block')
        await _session.block_user(self, unblock=False, client=client)


    async def unblock(self, *, from_client=None):
        """|coro|

        Unblock a user. Requires logged in client.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to unblock a user from. If ``None``,
            defaults to a client attached to this user.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to unblock a user.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'unblock')
        await _session.block_user(self, unblock=True, client=client)

    async def unfriend(self, *, from_client=None):
        """|coro|

        Try to unfriend a user. Requires logged in client.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to unfriend a user from. If ``None``,
            defaults to a client attached to this user.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to unfriend a user.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'unfriend')
        await _session.unfriend_user(self, client=client)

    async def send_friend_request(self, message: str = None, *, from_client=None):
        """|coro|

        Send a friend request to a user.

        .. note::

            This function does not raise any error if request was already sent.

        Parameters
        ----------
        message: :class:`str`
            A message to attach to a request.

        from_client: :class:`.Client`
            A logged in client to send request from. If ``None``,
            defaults to a client attached to this user.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to send a friend request to user.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'send_friend_request')
        await _session.send_friend_request(target=self, message=message, client=client)

    async def get_levels_on_page(self, page: int = 0, *, raise_errors: bool = True):
        """|coro|

        Fetches user's levels on a given page.

        This function is equivalent to calling:

        .. code-block:: python3

            await self._client.search_levels_on_page(
                page=page, filters=gd.Filters.setup_by_user(),
                user=self, raise_errors=raise_errors
            )
            # 'self' is an AbstractUser instance here.

        Parameters
        ----------
        page: :class:`int`
            Page to look for levels at.

        raise_errors: :class:`bool`
            Whether to raise errors if nothing was found or fetching failed.

        Returns
        -------
        List[:class:`.Level`]
            All levels found. Can be an empty list.
        """
        filters = Filters.setup_by_user()
        return await self._client.search_levels_on_page(
            page=page, filters=filters, user=self, raise_errors=raise_errors)

    async def get_levels(self, pages: Sequence[int] = None):
        """|coro|

        Gets levels on specified pages.

        This is equivalent to calling:

        .. code-block:: python3

            return await self._client.search_levels(
                pages=pages, filters=gd.Filters.setup_by_user(), user=self
            )
            # where 'self' is an AbstractUser instance.

        Parameters
        ----------
        pages: Sequence[:class:`int`]
            Pages to look at, represented as a finite sequence, so iterations can be performed.

        Returns
        -------
        List[:class:`.Level`]
            List of levels found. Can be an empty list.
        """
        filters = Filters.setup_by_user()
        return await self._client.search_levels(pages=pages, filters=filters, user=self)

    async def get_page_comments(self, page: int = 0):
        """|coro|

        Gets user's profile comments on a specific page.

        This is equivalent to:

        .. code-block:: python3

            await self.retrieve_page_comments('profile', page)
        """
        return await self.retrieve_page_comments('profile', page)

    async def get_page_comment_history(self,
        strategy: Union[int, str, CommentStrategy] = 0, page: int = 0):
        """|coro|

        Retrieves user's level comments. (history)

        Equivalent to calling:

        .. code-block:: python3

            await self.retrieve_page_comments('profile', page, strategy=strategy)
        """
        return await self.retrieve_page_comments('level', page, strategy=strategy)

    async def get_comments(self, pages: Sequence[int] = None):
        """|coro|

        Gets user's profile comments on specific pages.

        This is equivalent to the following:

        .. code-block:: python3

            await self.retrieve_comments('profile', pages)
        """
        if pages is None:
            pages = range(10)

        return await self.retrieve_comments('profile', pages)

    async def get_comment_history(
        self, strategy: Union[int, str, CommentStrategy] = 0, pages: Sequence[int] = None
    ):
        """|coro|

        Gets user's level (history) comments on specific pages.

        This is equivalent to the following:

        .. code-block:: python3

            await self.retrieve_comments('level', pages, strategy=strategy)
        """
        if pages is None:
            pages = range(10)

        return await self.retrieve_comments('level', pages, strategy=strategy)

    async def retrieve_page_comments(
        self, type: str = 'profile', page: int = 0, *, raise_errors: bool = True,
        strategy: Union[int, str, CommentStrategy] = 0
    ):
        """|coro|

        Utilizes getting comments. This is used in two other methods,
        :meth:`.User.get_page_comments` and :meth:`.User.get_page_comment_history`.

        Parameters
        ----------
        type: :class:`str`
            Type of comments to retrieve. Either `'profile'` or `'level'`.
            Defaults to `'profile'`.

        page: :class:`int`
            Page to look comments at.

        raise_errors: :class:`bool`
            Indicates whether :exc:`.NothingFound` should be raised.
            Should be set to false when getting several pages of comments,
            like in :meth:`.User.retrieve_comments`.

        strategy: Union[:class:`int`, :class:`str`, :class:`.CommentStrategy`]
            A strategy to apply when searching. This is converted to :class:`.CommentStrategy`
            using :func:`.utils.value_to_enum`.

        Returns
        -------
        List[:class:`.Comment`]
            List of all comments retrieved, if comments were found.

        Raises
        ------
        :exc:`.NothingFound`
            No comments were found.
        """
        strategy = value_to_enum(CommentStrategy, strategy)
        return await _session.retrieve_page_comments(
            type=type, user=self, page=page, raise_errors=raise_errors, strategy=strategy
        )

    async def retrieve_comments(
        self, type: str = 'profile', pages: Sequence[int] = None,
        strategy: Union[int, str, CommentStrategy] = 0
    ):
        """|coro|

        Utilizes getting comments on specified pages.

        Parameters
        ----------
        type: :class:`str`
            Type of comments to retrieve. Either `'profile'` or `'level'`.
            Defaults to `'profile'`.

        pages: Sequence[:class:`int`]
            Pages to look at, represented as a finite sequence, so iterations can be performed.

        strategy: Union[:class:`int`, :class:`str`, :class:`.CommentStrategy`]
            A strategy to apply when searching. This is converted to :class:`.CommentStrategy`
            using :func:`.utils.value_to_enum`.

        Returns
        -------
        List[:class:`.Comment`]
            List of comments found. Can be an empty list.
        """
        if pages is None:
            pages = range(10)

        strategy = value_to_enum(CommentStrategy, strategy)

        return await _session.retrieve_comments(
            type=type, user=self, pages=pages, strategy=strategy
        )



class XORCipher:

    @classmethod
    def cipher(cls, key: str, string: str):
        """Ciphers a string with XOR using key given.

        Due to the fact that *XOR* ``^`` operation is being used,
        the following is true:

        .. code-block:: python3

            xor = XORCipher
            xor.cipher('93582', xor.cipher('93582', 'NeKit')) == 'NeKit'  # True

        Parameters
        ----------
        key: :class:`str`
            A key to XOR cipher with, e.g. ``'59361'``.

        string: :class:`str`
            A string to apply XOR on.

        Returns
        -------
        :class:`str`
            A string after XOR operation.
        """
        keyB = tuple(map(ord, key))  # we need an index
        stringB = map(ord, string)
        result = ''

        for i, byte in enumerate(stringB):
            key_i = i % len(key)
            result += chr(byte ^ keyB[key_i])

        return result



import colorsys

from .utils.wrap_tools import make_repr
from .utils.converter import Converter

class Colour:
    """Represents a Colour. This class is similar
    to a (red, green, blue) :class:`tuple`.

    There is an alias for this called Color.

    Attributes
    ------------
    value: :class:`int`
        The raw integer colour value.
    """

    def __init__(self, value: int):
        if not isinstance(value, int):
            raise TypeError('Expected int parameter, but received {!r}.'.format(value.__class__.__name__))

        self.value = value

    def _get_byte(self, byte):
        return (self.value >> (8 * byte)) & 0xff

    def __eq__(self, other):
        return isinstance(other, Colour) and self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.to_hex()

    def __repr__(self):
        idx = self.index
        info = {
            'hex': self.to_hex(),
            'value': self.value,
            'index': idx if idx is None else Converter.to_ordinal(idx)
        }
        return make_repr(self, info)

    def __hash__(self):
        return hash(self.value)

    @property
    def index(self):
        """:class:`int`: Returns index that represents position of the colour in ``colors``.
        ``None`` if the colour is not present in ``colors``.
        """
        try:
            return colors.index(self)
        except ValueError:
            pass

    @property
    def r(self):
        """:class:`int`: Returns the red component of the colour."""
        return self._get_byte(2)

    @property
    def g(self):
        """:class:`int`: Returns the green component of the colour."""
        return self._get_byte(1)

    @property
    def b(self):
        """:class:`int`: Returns the blue component of the colour."""
        return self._get_byte(0)

    def to_hex(self):
        """:class:`str`: Returns the colour in hex format."""
        return '#{:0>6x}'.format(self.value)

    def to_rgb(self):
        """Tuple[:class:`int`, :class:`int`, :class:`int`]: Returns an (r, g, b) tuple representing the colour."""
        return (self.r, self.g, self.b)

    def to_rgba(self):
        """Tuple[:class:`int`, :class:`int`, :class:`int`, :class:`int`]: Same as :meth:`.Colour.to_rgb`, but contains alpha component. (always 255)"""
        return (*self.to_rgb(), 255)

    @classmethod
    def from_rgb(cls, r, g, b):
        """Constructs a :class:`Colour` from an RGB tuple."""
        return cls((r << 16) + (g << 8) + b)

    @classmethod
    def from_hsv(cls, h, s, v):
        """Constructs a :class:`Colour` from an HSV tuple."""
        rgb = colorsys.hsv_to_rgb(h, s, v)
        return cls.from_rgb(*(int(x * 255) for x in rgb))

Color = Colour

colors = (
    Colour(0x7dff00), Colour(0x00ff00), Colour(0x00ff7d),
    Colour(0x00ffff), Colour(0x007dff), Colour(0x0000ff),
    Colour(0x7d00ff), Colour(0xff00ff), Colour(0xff007d),
    Colour(0xff0000), Colour(0xff7d00), Colour(0xffff00),
    Colour(0xffffff), Colour(0xb900ff), Colour(0xffb900),
    Colour(0x000000), Colour(0x00c8ff), Colour(0xafafaf),
    Colour(0x5a5a5a), Colour(0xff7d7d), Colour(0x00af4b),
    Colour(0x007d7d), Colour(0x004baf), Colour(0x4b00af),
    Colour(0x7d007d), Colour(0xaf004b), Colour(0xaf4b00),
    Colour(0x7d7d00), Colour(0x4baf00), Colour(0xff4b00),
    Colour(0x963200), Colour(0x966400), Colour(0x649600),
    Colour(0x009664), Colour(0x006496), Colour(0x640096),
    Colour(0x960064), Colour(0x960000), Colour(0x009600),
    Colour(0x000096), Colour(0x7dffaf), Colour(0x7d7daf)
)
